# SPDX-FileCopyrightText: 2025-2026 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import typing as t
from collections import defaultdict
from pathlib import Path

from idf_build_apps import find_apps
from idf_build_apps.app import App
from idf_build_apps.args import FindArguments
from idf_build_apps.constants import ALL_TARGETS, BuildStatus
from jinja2 import Environment, FileSystemLoader

from idf_ci import get_pytest_cases
from idf_ci.idf_pytest import PytestCase

logger = logging.getLogger(__name__)


def add_test_case(test_cases: t.List[PytestCase], new_test_case: PytestCase) -> None:
    """Add a test case to the list if it does not exist."""
    if new_test_case not in test_cases:
        test_cases.append(new_test_case)


def group_test_cases_by_path(
    test_cases: t.List[PytestCase],
) -> t.Dict[str, t.Dict[str, t.Dict[str, t.List[PytestCase]]]]:
    """Group test cases by their path, target and sdkconfig."""
    # Structure: path -> target -> sdkconfig -> list[PytestCase]
    result: t.Dict[str, t.Dict[str, t.Dict[str, t.List[PytestCase]]]] = defaultdict(
        lambda: defaultdict(lambda: defaultdict(list))
    )

    for case in test_cases:
        path = Path(case.path).parent.as_posix()
        for target, config in zip(case.targets, case.configs):
            add_test_case(result[path][target][config], case)

    return result


def group_test_cases_by_target_and_config(
    test_cases: t.List[PytestCase],
) -> t.Dict[str, t.Dict[str, t.List[PytestCase]]]:
    """Group test cases by their target and sdkconfig."""
    # Structure: target -> sdkconfig -> list[PytestCase]
    result: t.Dict[str, t.Dict[str, t.List[PytestCase]]] = defaultdict(lambda: defaultdict(list))

    for case in test_cases:
        for target, config in zip(case.targets, case.configs):
            add_test_case(result[target][config], case)

    return result


def collect_apps(paths: t.List[str], include_only_enabled_apps: bool) -> t.Dict[str, t.Any]:
    """Collect all applications and corresponding test cases."""
    apps = find_apps(
        find_arguments=FindArguments(
            paths=paths,
            include_all_apps=not include_only_enabled_apps,
            recursive=True,
            enable_preview_targets=True,
        ),
    )

    test_cases = get_pytest_cases(
        paths=paths,
        marker_expr='',  # pass empty marker to collect all test cases
        additional_args=['--ignore-no-tests-collected-error'],
    )

    # Group apps by path
    apps_by_path: t.Dict[str, t.List[App]] = defaultdict(list)
    apps_by_abs_path: t.Dict[str, t.List[App]] = defaultdict(list)
    for app in apps:
        apps_by_path[app.app_dir].append(app)
        apps_by_abs_path[Path(app.app_dir).absolute().as_posix()].append(app)

    # Create a dict with test cases for quick lookup
    test_cases_index = group_test_cases_by_path(test_cases)

    result: t.Dict[str, t.Any] = {
        'summary': {
            'total_projects': len(apps_by_path),
            'total_apps': len(apps),
            'total_test_cases': len(test_cases),
            'total_test_cases_used': 0,
            'total_test_cases_disabled': 0,
            'total_test_cases_requiring_nonexistent_app': 0,
        },
        'projects': {},
    }

    for index, app_path in enumerate(sorted(apps_by_path)):
        logger.debug(
            f'Processing path {index + 1}/{len(apps_by_path)} with {len(apps_by_path[app_path])} apps: {app_path}'
        )

        result['projects'][app_path] = {'apps': [], 'test_cases_requiring_nonexistent_app': []}
        project: t.Dict[str, t.Any] = result['projects'][app_path]
        project_test_cases: t.Dict[str, PytestCase] = {}
        used_test_cases: t.Set[str] = set()
        disabled_test_cases: t.Set[str] = set()
        app_abs_path = Path(app_path).absolute().as_posix()

        for app in apps_by_path[app_path]:
            # Find test cases for current app by path, target and sdkconfig
            app_test_cases: t.Dict[str, PytestCase] = {}

            if app_abs_path in test_cases_index:
                # Gather all test cases
                for target_key in test_cases_index[app_abs_path]:
                    for sdkconfig_key in test_cases_index[app_abs_path][target_key]:
                        test_cases = test_cases_index[app_abs_path][target_key][sdkconfig_key]

                        for case in test_cases:
                            project_test_cases[case.caseid] = case

                # Find matching test cases
                if app.target in test_cases_index[app_abs_path]:
                    if app.config_name in test_cases_index[app_abs_path][app.target]:
                        test_cases = test_cases_index[app_abs_path][app.target][app.config_name]

                        for case in test_cases:
                            app_test_cases[case.caseid] = case

            # Get enabled test targets from manifest if exists
            enabled_test_targets = ALL_TARGETS
            if app.MANIFEST is not None:
                enabled_test_targets = app.MANIFEST.enable_test_targets(app_path, config_name=app.config_name)

            # Test cases info
            test_cases_info: t.List[t.Dict[str, t.Any]] = []
            for case in app_test_cases.values():
                test_case: t.Dict[str, t.Any] = {
                    'name': case.name,
                    'caseid': case.caseid,
                }

                skipped_targets = case.skipped_targets()
                is_disabled_by_manifest = app.target not in enabled_test_targets
                is_disabled_by_marker = app.target in skipped_targets

                if is_disabled_by_manifest or is_disabled_by_marker:
                    test_case['disabled'] = True
                    test_case['disabled_by_manifest'] = is_disabled_by_manifest
                    test_case['disabled_by_marker'] = is_disabled_by_marker

                    if is_disabled_by_marker:
                        test_case['skip_reason'] = skipped_targets[app.target]

                    if is_disabled_by_manifest:
                        test_case['test_comment'] = app.test_comment or ''

                    disabled_test_cases.add(case.caseid)
                else:
                    used_test_cases.add(case.caseid)

                test_cases_info.append(test_case)

            project['apps'].append(
                {
                    'target': app.target,
                    'sdkconfig': app.config_name,
                    'build_status': app.build_status.value,
                    'build_comment': app.build_comment or '',
                    'test_comment': app.test_comment or '',
                    'test_cases': test_cases_info,
                }
            )

        unused_test_cases: t.Set[str] = set(project_test_cases.keys())
        unused_test_cases.difference_update(used_test_cases)
        unused_test_cases.difference_update(disabled_test_cases)

        # Group unused test cases by target and sdkconfig
        unused_test_cases_grouped = group_test_cases_by_target_and_config(
            [project_test_cases[caseid] for caseid in unused_test_cases]
        )

        # Add unused test cases info to the project
        for target in unused_test_cases_grouped:
            for sdkconfig in unused_test_cases_grouped[target]:
                project['test_cases_requiring_nonexistent_app'].append(
                    {
                        'target': target,
                        'sdkconfig': sdkconfig,
                        'test_cases': [
                            {
                                'name': case.name,
                                'caseid': case.caseid,
                            }
                            for case in unused_test_cases_grouped[target][sdkconfig]
                        ],
                    }
                )

        result['summary']['total_test_cases_used'] += len(used_test_cases)
        result['summary']['total_test_cases_disabled'] += len(disabled_test_cases)
        result['summary']['total_test_cases_requiring_nonexistent_app'] += len(unused_test_cases)

    return result


def format_as_json(data: t.Dict[str, t.Any]) -> str:
    """Format collected data as JSON."""
    return json.dumps(data)


def format_as_html(data: t.Dict[str, t.Any]) -> str:
    """Format collected data as HTML."""
    rows = []
    projects = data.get('projects', {})
    all_target_list = set()

    for project_path, project_info in projects.items():
        apps = project_info.get('apps', [])

        total_tests = 0
        total_enabled_tests = 0
        target_list = set()
        config_list = set()
        config_target_app: t.Dict[str, t.Dict[str, t.Any]] = defaultdict(dict)
        details = []

        for app in apps:
            # Collect sdkconfigs
            config_list.add(app.get('sdkconfig'))

            # Collect targets
            target_list.add(app.get('target'))

            # config -> target -> app
            config_target_app[app.get('sdkconfig')][app.get('target')] = app

        all_target_list.update(target_list)
        config_list_sorted = sorted(config_list)
        target_list_sorted = sorted(target_list)

        for config in config_list_sorted:
            detail_item = {'sdkconfig': config, 'coverage': 0, 'targets': []}

            targets_tested = 0
            targets_total = 0

            for target in target_list_sorted:
                # Status:
                # U - Unknown
                # B - Should be built
                # D - Disabled
                # T - Tests enabled
                # S - Tests skipped
                target_info = {
                    'name': target,
                    'status': 'U',
                    'status_label': '',  # B - Should be built, D - Disabled
                    'has_err': False,
                    'is_disabled': False,
                    'disable_reason': '',
                    'tests': 0,
                    'enabled_tests': 0,
                }

                app = config_target_app.get(config, {}).get(target)

                if app:
                    targets_total += 1

                    status = app.get('build_status')
                    test_cases = app.get('test_cases', [])
                    disabled_test_cases = [case for case in test_cases if case.get('disabled')]

                    target_info['tests'] = len(test_cases)
                    target_info['enabled_tests'] = target_info['tests'] - len(disabled_test_cases)

                    total_tests += target_info['tests']
                    total_enabled_tests += target_info['enabled_tests']

                    if status == BuildStatus.DISABLED:
                        target_info['status'] = target_info['status_label'] = 'D'
                        target_info['is_disabled'] = True
                        target_info['disable_reason'] = app.get('build_comment', '')
                    elif status == BuildStatus.SHOULD_BE_BUILT:
                        target_info['status'] = target_info['status_label'] = 'B'

                    if target_info['enabled_tests'] > 0:
                        target_info['status'] += 'T'
                        targets_tested += 1

                    if len(disabled_test_cases) > 0:
                        target_info['status'] += 'S'

                    # Mismatched test cases
                    disabled_by_manifest_only = [
                        case
                        for case in test_cases
                        if case.get('disabled_by_manifest') and not case.get('disabled_by_marker')
                    ]
                    disabled_by_marker_only = [
                        case
                        for case in test_cases
                        if case.get('disabled_by_marker') and not case.get('disabled_by_manifest')
                    ]

                    # Set error if there are any mismatches
                    if disabled_by_manifest_only or disabled_by_marker_only:
                        target_info['has_err'] = True
                        target_info['disabled_by_manifest_only'] = disabled_by_manifest_only
                        target_info['disabled_by_marker_only'] = disabled_by_marker_only

                detail_item['targets'].append(target_info)

            if targets_total > 0:
                detail_item['coverage'] = (targets_tested / targets_total) * 100

            details.append(detail_item)

        tests_unknown_sdkconfig = []
        for item in project_info.get('test_cases_requiring_nonexistent_app', []):
            for test_case in item.get('test_cases', []):
                caseid = test_case.get('caseid')
                if caseid is not None:
                    tests_unknown_sdkconfig.append(caseid)

        rows.append(
            {
                'project_path': project_path,
                'apps': len(apps),
                'tests': total_tests,
                'enabled_tests': total_enabled_tests,
                'tests_unknown_sdkconfig': tests_unknown_sdkconfig,
                'target_list': target_list_sorted,
                'details': details,
            }
        )

    rows = sorted(rows, key=lambda x: x['project_path'])
    loader = FileSystemLoader(Path(__file__).parent)
    env = Environment(loader=loader)
    template = env.get_template('template.html')
    output = template.render(
        {
            'targets': sorted(all_target_list),
            'rows': rows,
        }
    )

    return output



# SPDX-FileCopyrightText: 2025-2026 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import inspect
import typing as t
from pathlib import Path

import click
from pydantic import BaseModel
from tomlkit import TOMLDocument, load
from tomlkit import dumps as toml_dumps

from idf_ci.settings import CiSettings, get_ci_settings, pick_toml_file


def _parse_config_key(config_key: str) -> t.List[str]:
    """Parse a dot-separated path into a list of parts."""
    parts = [p.strip() for p in config_key.split('.') if p.strip()]
    if not parts:
        raise click.BadParameter('Empty path')
    return parts


def _iter_config_key_paths(data: t.Mapping[str, t.Any], prefix: str = '') -> t.Iterable[str]:
    for key in sorted(data.keys()):
        path = f'{prefix}.{key}' if prefix else key
        yield path
        value = data[key]
        if isinstance(value, dict):
            yield from _iter_config_key_paths(value, path)


def _complete_config_key(ctx, param, incomplete: str) -> t.List[str]:  # noqa: ARG001
    settings = get_ci_settings()
    data = settings.model_dump(mode='python', exclude_none=True)
    keys = _iter_config_key_paths(data)
    return [key for key in keys if key.startswith(incomplete)]


def _format_type(annotation: t.Any) -> str:
    origin = t.get_origin(annotation)
    if origin is None:
        if isinstance(annotation, type):
            return annotation.__name__
        return repr(annotation)

    if origin is t.Literal:
        return f'Literal[{", ".join(repr(a) for a in t.get_args(annotation))}]'

    if origin is t.Union:
        return f'Union[{", ".join(_format_type(a) for a in t.get_args(annotation))}]'

    return f'{origin.__name__}[{", ".join(_format_type(a) for a in t.get_args(annotation))}]'


def _get_model_class(annotation: t.Any) -> t.Optional[t.Type[BaseModel]]:
    if inspect.isclass(annotation) and issubclass(annotation, BaseModel):
        return annotation

    origin = t.get_origin(annotation)
    if origin is t.Union:
        for arg in t.get_args(annotation):
            if inspect.isclass(arg) and issubclass(arg, BaseModel):
                return arg

    return None


def _resolve_field(config_key: str) -> t.Tuple[t.Optional[t.Type[BaseModel]], t.Any, str, t.Type[BaseModel]]:
    """Resolve a config key path to find the field and its containing class."""
    parts = _parse_config_key(config_key)

    current_cls: t.Optional[t.Type[BaseModel]] = CiSettings
    parent_cls: t.Type[BaseModel] = CiSettings
    field = None
    for part in parts:
        if current_cls is None:
            raise click.BadParameter(f'Config key is not a section: {".".join(parts)}')
        if part not in current_cls.model_fields:
            raise click.BadParameter(f'Unknown config key: {config_key}')
        parent_cls = current_cls
        field = current_cls.model_fields[part]
        current_cls = _get_model_class(field.annotation)

    return current_cls, field, parts[-1], parent_cls


def argument_config_key(func):
    return click.argument('config_key', shell_complete=_complete_config_key)(func)


@click.group()
def config():
    """Inspect and manage idf-ci configuration."""
    pass


@config.command()
@argument_config_key
def show(config_key: str):
    """Show the effective configuration after all overrides are applied."""
    settings = get_ci_settings()
    data = settings.model_dump(mode='python', exclude_none=True)
    config_path = _get_config_file_path()
    config_label = str(config_path) if config_path else '<not found>'
    config_data: t.Optional[TOMLDocument] = None
    if config_path:
        with open(config_path) as f:
            config_data = load(f)

    value = _get_value_by_config_key(data, config_key)
    # Determine value source
    if (
        CiSettings.CLI_OVERRIDES
        and _get_value_by_config_key(CiSettings.CLI_OVERRIDES, config_key, safe=True) is not None
    ):
        source = 'cli_override'
    elif config_data and _get_value_by_config_key(config_data, config_key, safe=True) is not None:
        source = 'config_file'
    else:
        source = 'default'

    # Show config file location
    click.echo(f'Config file: {config_label}')

    # Show value source
    if source == 'cli_override':
        click.echo('Source: CLI override (--config option)')
    elif source == 'config_file':
        click.echo(f'Source: Config file ({config_label})')
    else:
        click.echo('Source: Default value')

    click.echo()
    click.echo(_format_toml_value(config_key, value))


@config.command()
@argument_config_key
def explain(config_key: str):
    """Explain a config key, including type, default, and description."""
    next_cls, field, _, _ = _resolve_field(config_key)

    click.echo(f'Key: {config_key}')
    click.echo(f'Type: {_format_type(field.annotation)}')
    if field.description:
        click.echo(f'Description: {field.description}')

    if next_cls is not None and issubclass(next_cls, BaseModel):
        subkeys = sorted(next_cls.model_fields.keys())
        if subkeys:
            click.echo('Subkeys:')
            for key in subkeys:
                click.echo(f'  - {config_key}.{key}')
    else:
        if field.is_required():
            click.echo('Default: <required>')
        else:
            click.echo('Default (toml):')
            click.echo('```toml')
            click.echo(_format_toml_value(config_key, field.default))
            click.echo('```')

    origin = t.get_origin(field.annotation)
    if origin is t.Literal:
        values = ', '.join(repr(v) for v in t.get_args(field.annotation))
        click.echo(f'Choices: {values}')


def _get_value_by_config_key(data: t.Dict[str, t.Any], config_key: str, *, safe: bool = False) -> t.Any:
    """Navigate nested dict using dot-separated config key path.

    :param data: The dictionary to navigate
    :param config_key: Dot-separated path to the value
    :param safe: If True, return None on error instead of raising exceptions

    :returns: The value at the path, or None if safe=True and path is invalid
    """
    parts = _parse_config_key(config_key)
    cursor: t.Any = data
    for part in parts:
        if not isinstance(cursor, dict) or part not in cursor:
            if safe:
                return None
            raise click.BadParameter(f'Unknown config key: {config_key}')
        cursor = cursor[part]
    return cursor


def _get_config_file_path() -> t.Optional[Path]:
    """Get the path to the config file being used."""
    return pick_toml_file(CiSettings.CONFIG_FILE_PATH)


def _format_toml_value(path: str, value: t.Any) -> str:
    """Format a value as TOML syntax with proper nesting for the given path."""
    parts = _parse_config_key(path)

    if isinstance(value, str):
        # Use TOML multiline string format if the value contains newlines
        # This preserves newlines for CLI readability while being TOML-valid
        if '\n' in value or '\r' in value:
            # Escape triple quotes and backslashes for multiline strings
            escaped = value.replace('\\', '\\\\').replace('"""', '\\"""')
            return f'{path} = """\n{escaped}\n"""'
        # For single-line strings, escape special characters
        escaped = value.replace('\\', '\\\\').replace('"', '\\"')
        return f'{path} = "{escaped}"'

    # For complex values, build nested structure and let toml_dumps handle formatting
    data: t.Dict[str, t.Any] = {}
    cursor = data
    for part in parts[:-1]:
        cursor[part] = {}
        cursor = cursor[part]
    cursor[parts[-1]] = value
    return toml_dumps(data).strip()



# SPDX-FileCopyrightText: 2025-2026 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import glob
import json
import logging
import os
import re
import subprocess
import tempfile
import time
import typing as t
import zipfile
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from datetime import timedelta
from functools import lru_cache
from pathlib import Path

import esp_bool_parser
import minio
import requests
import urllib3
from gitlab import Gitlab
from minio import Minio

from .._compat import UNDEF, is_undefined
from .._vendor import translate
from ..envs import GitlabEnvVars
from ..settings import get_ci_settings
from ..utils import get_current_branch

logger = logging.getLogger(__name__)


def execute_concurrent_tasks(
    tasks: t.List[t.Callable[..., t.Any]],
    max_workers: t.Optional[int] = None,
    task_name: str = 'executing task',
) -> t.List[t.Any]:
    """Execute tasks concurrently using ThreadPoolExecutor.

    :param tasks: List of callable tasks to execute
    :param max_workers: Maximum number of worker threads
    :param task_name: Error message prefix for logging

    :returns: List of successful task results; order is not guaranteed
    """
    results = []
    errors = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(task) for task in tasks]

        for future in as_completed(futures):
            try:
                result = future.result()
                if result is not None:
                    results.append(result)
            except Exception as e:
                logger.error(f'Error while {task_name}: {e}')
                errors.append(e)

    if errors:
        _nl = '\n'  # compatible with Python < 3.12
        raise ArtifactError(f'Got {len(errors)} errors while {task_name}:\n{_nl.join([f"- {e}" for e in errors])}')

    return results


@dataclass
class ArtifactParams:
    """Common parameters for artifacts operations.

    The commit SHA can be determined in the following order of precedence:

    1. Explicitly provided commit_sha parameter
    2. PIPELINE_COMMIT_SHA environment variable
    3. Latest commit from branch (where branch is determined by branch parameter or
       current git branch)
    """

    commit_sha: t.Optional[str] = None
    branch: t.Optional[str] = None
    folder: t.Optional[str] = None

    def __post_init__(self):
        if self.folder is None:
            self.folder = os.getcwd()
        self.from_path = Path(self.folder)

        # Get commit SHA with the following precedence:
        # 1. CLI provided commit_sha
        if self.commit_sha:
            return

        # 2. Environment variable PIPELINE_COMMIT_SHA
        if os.getenv('PIPELINE_COMMIT_SHA'):
            self.commit_sha = os.environ['PIPELINE_COMMIT_SHA']
            return

        # 3. Latest commit from branch
        try:
            if self.branch is None:
                self.branch = get_current_branch()
            result = subprocess.run(
                ['git', 'rev-parse', self.branch],
                check=True,
                capture_output=True,
                encoding='utf-8',
            )
            self.commit_sha = result.stdout.strip()
        except Exception:
            raise ValueError(
                'Failed to get commit SHA from git command. '
                'Must set commit_sha or branch parameter, or set PIPELINE_COMMIT_SHA env var'
            )


class ArtifactError(RuntimeError):
    """Base exception for artifact-related errors."""


class S3Error(ArtifactError):
    """Exception raised for S3-related errors."""


class PresignedUrlError(ArtifactError):
    """Exception raised for presigned URL-related errors."""


class ArtifactManager:
    def __init__(self):
        self.envs = GitlabEnvVars()
        self.settings = get_ci_settings()

        self._s3_client: t.Optional[Minio] = UNDEF  # type: ignore
        self._s3_public_client: t.Optional[Minio] = UNDEF  # type: ignore

    @property
    @lru_cache()
    def gl(self):
        return Gitlab(
            self.envs.GITLAB_HTTPS_SERVER,
            private_token=self.envs.GITLAB_ACCESS_TOKEN,
        )

    @property
    @lru_cache()
    def project(self):
        project = self.gl.projects.get(self.settings.gitlab.project)
        if not project:
            raise ValueError(f'Project {self.settings.gitlab.project} not found')
        return project

    @property
    def s3_client(self) -> t.Optional[minio.Minio]:
        if is_undefined(self._s3_client):
            self._s3_client = self._create_s3_client()
        return self._s3_client

    @property
    def s3_public_client(self) -> t.Optional[minio.Minio]:
        if is_undefined(self._s3_public_client):
            self._s3_public_client = self._create_s3_client(public=True)
        return self._s3_public_client

    def _create_s3_client(self, *, public=False) -> t.Optional[minio.Minio]:
        if not self.envs.IDF_S3_SERVER:
            logger.info('S3 credentials not available. Skipping S3 features...')
            return None

        if not public:
            if not all(
                [
                    self.envs.IDF_S3_ACCESS_KEY,
                    self.envs.IDF_S3_SECRET_KEY,
                ]
            ):
                logger.info('S3 credentials not available. Skipping S3 features...')
                return None

        if self.envs.IDF_S3_SERVER.startswith('https://'):
            host = self.envs.IDF_S3_SERVER.replace('https://', '')
            secure = True
        elif self.envs.IDF_S3_SERVER.startswith('http://'):
            host = self.envs.IDF_S3_SERVER.replace('http://', '')
            secure = False
        else:
            raise ValueError('Please provide a http or https server URL for S3')

        logger.debug('S3 Host: %s', host)
        return minio.Minio(
            host,
            access_key=self.envs.IDF_S3_ACCESS_KEY if not public else '',
            secret_key=self.envs.IDF_S3_SECRET_KEY if not public else '',
            secure=secure,
            http_client=urllib3.PoolManager(
                num_pools=10,
                maxsize=10,
                timeout=urllib3.Timeout(
                    total=self.envs.IDF_S3_TIMEOUT_TOTAL,
                ),
                retries=urllib3.Retry(
                    total=5,
                    backoff_factor=1,
                    status_forcelist=(408, 429, 500, 502, 503, 504),
                ),
            ),
        )

    def _get_patterns_for_type(self, artifact_type: str) -> t.List[str]:
        config = self.settings.gitlab.artifacts.s3.configs[artifact_type]

        if not config.base_dir_pattern:
            return config.patterns

        return [os.path.join(config.base_dir_pattern, pattern) for pattern in config.patterns]

    def _get_artifact_types(self, artifact_type: t.Optional[str]) -> t.List[str]:
        if artifact_type and artifact_type not in self.settings.gitlab.artifacts.s3.configs:
            raise ValueError(
                f'Invalid artifact type: {artifact_type}. '
                f'Available types: {list(self.settings.gitlab.artifacts.s3.configs.keys())}'
            )

        available_types = []
        for art_type, config in self.settings.gitlab.artifacts.s3.configs.items():
            # Check if_clause condition
            if config.if_clause:
                try:
                    stmt = esp_bool_parser.parse_bool_expr(config.if_clause)
                    res = stmt.get_value('', '')
                except Exception as e:
                    logger.debug(
                        f'Skipping {art_type} artifacts due to error '
                        f'while evaluating if_clause: {config.if_clause}: {e}'
                    )
                else:
                    if res:
                        available_types.append(art_type)
                    else:
                        logger.debug(f'Skipping {art_type} artifacts due to if_clause: {config.if_clause}')
            else:
                available_types.append(art_type)

        if artifact_type:
            if artifact_type in available_types:
                return [artifact_type]
            else:
                return []

        return available_types

    def _get_s3_path(self, prefix: str, from_path: Path) -> str:
        if from_path.is_absolute():
            # Resolve IDF_PATH to absolute path to ensure relative_to() works correctly
            idf_path = Path(self.envs.IDF_PATH).resolve() if self.envs.IDF_PATH else Path.cwd()
            rel_path = str(from_path.relative_to(idf_path))
        else:
            rel_path = str(from_path)

        return f'{prefix}{rel_path}' if rel_path != '.' else prefix

    def _download_files_from_s3(
        self,
        *,
        prefix: str,
        from_path: Path,
        artifact_type: str,
    ) -> int:
        config = self.settings.gitlab.artifacts.s3.configs[artifact_type]
        s3_client = self._validate_s3_client(artifact_type)

        def _download_task(_obj_name: str, _output_path: Path) -> None:
            logger.debug(f'Downloading {_obj_name} to {_output_path}')
            _output_path.parent.mkdir(parents=True, exist_ok=True)
            s3_client.fget_object(config.bucket, _obj_name, str(_output_path))

        tasks = []
        patterns_regexes = [
            re.compile(translate(pattern, recursive=True, include_hidden=True))
            for pattern in self._get_patterns_for_type(artifact_type)
        ]
        for obj in s3_client.list_objects(config.bucket, prefix=self._get_s3_path(prefix, from_path), recursive=True):
            output_path = Path(self.envs.IDF_PATH) / obj.object_name.replace(prefix, '')
            if not any(pattern.match(str(output_path)) for pattern in patterns_regexes):
                continue
            tasks.append(
                lambda _obj_name=obj.object_name, _output_path=output_path: _download_task(_obj_name, _output_path)
            )

        execute_concurrent_tasks(tasks, task_name='downloading object')
        return len(tasks)

    def _extract_zip_file(self, zip_path: Path) -> None:
        logger.debug(f'Extracting {zip_path}')
        try:
            with zipfile.ZipFile(zip_path, 'r') as zipf:
                zipf.extractall(zip_path.parent)
        except zipfile.BadZipFile as e:
            logger.error(f'Failed to extract {zip_path}: {e}')
            raise
        finally:
            zip_path.unlink()
            logger.debug(f'Removed zip file {zip_path}')

    def _download_zip_from_s3(
        self,
        *,
        prefix: str,
        from_path: Path,
        artifact_type: str,
    ) -> int:
        """Download and extract zip files from S3."""
        config = self.settings.gitlab.artifacts.s3.configs[artifact_type]
        s3_client = self._validate_s3_client(artifact_type)

        def _download_and_extract(obj_name: str, zip_path: Path) -> None:
            logger.debug(f'Downloading {obj_name} to {zip_path}')
            zip_path.parent.mkdir(parents=True, exist_ok=True)
            s3_client.fget_object(config.bucket, obj_name, str(zip_path))
            self._extract_zip_file(zip_path)

        # Find matching zip files
        tasks = []
        # Look for zip files matching artifact types (e.g., flash.zip, debug.zip)
        # Since we're listing recursively, just check if filename matches {art_type}.zip
        for obj in s3_client.list_objects(config.bucket, prefix=self._get_s3_path(prefix, from_path), recursive=True):
            output_path = Path(self.envs.IDF_PATH) / obj.object_name.replace(prefix, '')
            if output_path.name == f'{artifact_type}.zip':
                tasks.append(lambda o=obj.object_name, op=output_path: _download_and_extract(o, op))

        execute_concurrent_tasks(tasks, task_name='downloading and extracting zip')
        return len(tasks)

    def _validate_s3_client(self, artifact_type: str) -> minio.Minio:
        config = self.settings.gitlab.artifacts.s3.configs[artifact_type]
        if config.is_public:
            if not self.s3_public_client:
                raise S3Error('S3 public client is not configured properly')
            return self.s3_public_client
        else:
            if not self.s3_client:
                raise S3Error('S3 client is not configured properly')
            return self.s3_client

    def _upload_files_to_s3(
        self,
        *,
        prefix: str,
        from_path: Path,
        artifact_type: str,
    ) -> int:
        config = self.settings.gitlab.artifacts.s3.configs[artifact_type]
        s3_client = self._validate_s3_client(artifact_type)

        def _upload_task(_filepath: Path, _s3_path: str) -> None:
            logger.debug(f'Uploading {_filepath} to {_s3_path}')
            s3_client.fput_object(config.bucket, _s3_path, str(_filepath))

        tasks = []
        for pattern in config.patterns:
            if config.base_dir_pattern:
                pattern = os.path.join(config.base_dir_pattern, pattern)

            abs_pattern = os.path.join(str(from_path), pattern)
            for file_str in glob.glob(abs_pattern, recursive=True):
                filepath = Path(file_str)
                if not filepath.is_file():
                    continue

                s3_path = self._get_s3_path(prefix, filepath)
                tasks.append(lambda _filepath=filepath, _s3_path=s3_path: _upload_task(_filepath, _s3_path))

        execute_concurrent_tasks(tasks, task_name='uploading file')
        return len(tasks)

    def _upload_zip_to_s3(
        self,
        *,
        prefix: str,
        from_path: Path,
        artifact_type: str,
    ) -> int:
        """Upload artifacts as zip files to S3.

        This method:

        1. Finds directories matching ``base_dir_pattern`` (only folders).
        2. For each directory, finds files matching ``patterns`` (relative to that
           directory).
        3. Creates a zip file named ``<artifact_type>.zip`` in that directory.
        4. Uploads the zip file to S3.

        :param prefix: S3 prefix path
        :param from_path: Base path to search from
        :param artifact_type: Type of artifact (used as zip filename)

        :returns: Number of zip files uploaded
        """
        config = self.settings.gitlab.artifacts.s3.configs[artifact_type]
        s3_client = self._validate_s3_client(artifact_type)

        def _upload_zip_task(_zip_path: Path, _s3_path: str) -> None:
            logger.debug(f'Uploading zip {_zip_path} to {_s3_path}')
            s3_client.fput_object(config.bucket, _s3_path, str(_zip_path))

        tasks = []
        # Find all directories matching base_dir_pattern
        # The pattern should match directories only (e.g., '**/build*/')
        if not config.base_dir_pattern:
            matching_dirs = [from_path.resolve()]
        else:
            abs_pattern = os.path.realpath(str(Path(from_path) / config.base_dir_pattern))
            # Remove trailing slash if present for glob matching
            # Pattern like '**/build*/' should match directories starting with 'build'
            pattern_for_glob = abs_pattern.rstrip('/')

            # Use glob to find all matches, then filter for directories only
            matching_dirs = [
                Path(match).resolve() for match in glob.glob(pattern_for_glob, recursive=True) if Path(match).is_dir()
            ]

        logger.debug(f'Found {len(matching_dirs)} directories matching pattern {config.base_dir_pattern}')

        # For each matching directory, collect files and create zip
        for basedir in matching_dirs:
            files_to_zip = []
            # Search for files matching patterns relative to basedir
            for file_pattern in config.patterns:
                for fps in glob.glob(os.path.join(str(basedir), file_pattern), recursive=True):
                    if os.path.isfile(fps):
                        files_to_zip.append(Path(fps))

            if not files_to_zip:
                logger.debug(f'No files found in {basedir} matching patterns {config.patterns}')
                continue

            # Create zip file in the basedir with name <artifact_type>.zip
            zip_path = basedir / f'{artifact_type}.zip'
            logger.debug(f'Creating zip {zip_path} with {len(files_to_zip)} files')

            with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
                # Resolve basedir to absolute path to ensure relative_to() works correctly
                basedir_resolved = basedir.resolve()
                for fp in files_to_zip:
                    # Resolve filepath to absolute path before computing relative path
                    filepath_resolved = fp.resolve()
                    # Add file to zip with path relative to basedir
                    arcname = filepath_resolved.relative_to(basedir_resolved)
                    zipf.write(filepath_resolved, arcname)

            # Upload the zip file
            s3_path = self._get_s3_path(prefix, zip_path)
            tasks.append(lambda _zip_path=zip_path, _s3_path=s3_path: _upload_zip_task(_zip_path, _s3_path))

        execute_concurrent_tasks(tasks, task_name='uploading zip file')
        return len(tasks)

    def _download_files_from_presigned_json(
        self,
        presigned_json: str,
        from_path: Path,
        artifact_type: str,
    ) -> int:
        with open(presigned_json) as f:
            presigned_urls = json.load(f)

        from_path_rel = from_path.relative_to(self.envs.IDF_PATH)
        patterns = self._get_patterns_for_type(artifact_type)
        if from_path_rel != Path('.'):
            patterns = [os.path.join(str(from_path_rel), pattern) for pattern in patterns]
        patterns_regexes = [re.compile(translate(pattern, recursive=True, include_hidden=True)) for pattern in patterns]

        def _download_task(_url: str, _output_path: Path) -> None:
            logger.debug(f'Downloading {_url} to {_output_path}')
            _output_path.parent.mkdir(parents=True, exist_ok=True)

            response = requests.get(_url, stream=True)
            if response.status_code != 200:
                raise PresignedUrlError(f'Failed to download {_output_path.name}: {response.status_code}')

            with open(_output_path, 'wb') as f:
                f.write(response.content)

        tasks = []
        for rel_path, url in presigned_urls.items():
            if not any(pattern.match(str(rel_path)) for pattern in patterns_regexes):
                continue

            output_path = Path(self.envs.IDF_PATH) / rel_path
            tasks.append(lambda _url=url, _output_path=output_path: _download_task(_url, _output_path))

        execute_concurrent_tasks(tasks, task_name='downloading object')
        return len(tasks)

    def _download_zip_from_presigned_json(
        self,
        presigned_json: str,
        from_path: Path,
        artifact_type: str,
    ) -> int:
        """Download and extract zip files from presigned URLs."""
        with open(presigned_json) as f:
            presigned_urls = json.load(f)

        zip_filename = f'{artifact_type}.zip'
        from_path_rel = from_path.relative_to(self.envs.IDF_PATH)

        def _download_and_extract(url: str, output_path: Path) -> None:
            logger.debug(f'Downloading {url} to {output_path}')
            output_path.parent.mkdir(parents=True, exist_ok=True)

            response = requests.get(url, stream=True)
            if response.status_code != 200:
                raise PresignedUrlError(f'Failed to download {output_path.name}: {response.status_code}')

            with open(output_path, 'wb') as f:
                f.write(response.content)

            self._extract_zip_file(output_path)

        tasks = []
        for rel_path, url in presigned_urls.items():
            rel_path_obj = Path(rel_path)
            if rel_path_obj.name != zip_filename:
                continue

            if from_path_rel != Path('.') and from_path_rel not in rel_path_obj.parents:
                continue

            output_path = Path(self.envs.IDF_PATH) / rel_path
            tasks.append(lambda u=url, op=output_path: _download_and_extract(u, op))

        execute_concurrent_tasks(tasks, task_name='downloading and extracting zip from presigned URLs')
        return len(tasks)

    def download_artifacts(
        self,
        *,
        commit_sha: t.Optional[str] = None,
        branch: t.Optional[str] = None,
        artifact_type: t.Optional[str] = None,
        folder: t.Optional[str] = None,
        presigned_json: t.Optional[str] = None,
    ) -> None:
        """Download artifacts from S3 or via presigned URLs.

        When presigned_json is provided, downloads artifacts from presigned URLs.

        :param commit_sha: Optional commit SHA. If no commit_sha provided, will use

            1. PIPELINE_COMMIT_SHA env var
            2. latest commit from branch
        :param branch: Optional Git branch. If no branch provided, will use current
            branch
        :param artifact_type: Type of artifacts to download (debug, flash, metrics)
        :param folder: Download artifacts under this folder
        :param presigned_json: Path to the presigned.json file for download

        :raises ValueError: If S3 artifacts are not enabled
        """
        if not self.settings.gitlab.artifacts.s3.enable:
            raise ValueError('S3 artifacts are not enabled in the CI settings')

        params = ArtifactParams(
            commit_sha=commit_sha,
            branch=branch,
            folder=folder,
        )

        start_time = time.time()
        downloaded_count = 0

        if not presigned_json:
            # download from s3 directly
            logger.info(f'Downloading artifacts under {params.from_path} from s3 (commit sha: {params.commit_sha})')

            for art_type in self._get_artifact_types(artifact_type):
                config = self.settings.gitlab.artifacts.s3.configs[art_type]
                if config.zip_first:
                    downloaded_count += self._download_zip_from_s3(
                        prefix=f'{self.settings.gitlab.project}/{params.commit_sha}/',
                        from_path=params.from_path,
                        artifact_type=art_type,
                    )
                else:
                    downloaded_count += self._download_files_from_s3(
                        prefix=f'{self.settings.gitlab.project}/{params.commit_sha}/',
                        from_path=params.from_path,
                        artifact_type=art_type,
                    )
            logger.info(f'Downloaded {downloaded_count} artifacts in {time.time() - start_time:.2f} seconds')
            return

        # download from presigned urls
        logger.info(f'Downloading artifacts under {params.from_path} from presigned JSON')

        for art_type in self._get_artifact_types(artifact_type):
            config = self.settings.gitlab.artifacts.s3.configs[art_type]
            if config.zip_first:
                downloaded_count += self._download_zip_from_presigned_json(
                    presigned_json,
                    params.from_path,
                    art_type,
                )
            else:
                downloaded_count += self._download_files_from_presigned_json(
                    presigned_json,
                    params.from_path,
                    art_type,
                )
        logger.info(f'Downloaded {downloaded_count} artifacts in {time.time() - start_time:.2f} seconds')

    def upload_artifacts(
        self,
        *,
        commit_sha: t.Optional[str] = None,
        branch: t.Optional[str] = None,
        artifact_type: t.Optional[str] = None,
        folder: t.Optional[str] = None,
    ) -> None:
        """Upload artifacts to S3.

        :param commit_sha: Optional commit SHA. If no commit_sha provided, will use 1)
            PIPELINE_COMMIT_SHA env var, 2) latest commit from branch
        :param branch: Optional Git branch. If no branch provided, will use current
            branch
        :param artifact_type: Type of artifacts to upload (debug, flash, metrics)
        :param folder: Upload artifacts under this folder

        :raises ValueError: If S3 artifacts are not enabled
        :raises S3Error: If S3 is not configured
        """
        if not self.settings.gitlab.artifacts.s3.enable:
            raise ValueError('S3 artifacts are not enabled in the CI settings')

        params = ArtifactParams(
            commit_sha=commit_sha,
            branch=branch,
            folder=folder,
        )

        if not self.s3_client:
            raise S3Error('Configure S3 storage to upload artifacts')

        prefix = f'{self.settings.gitlab.project}/{params.commit_sha}/'
        logger.info(f'Uploading artifacts under {params.from_path} to s3 (commit sha: {params.commit_sha})')

        start_time = time.time()
        uploaded_count = 0

        for art_type in self._get_artifact_types(artifact_type):
            config = self.settings.gitlab.artifacts.s3.configs[art_type]
            if config.zip_first:
                uploaded_count += self._upload_zip_to_s3(
                    prefix=prefix,
                    from_path=params.from_path,
                    artifact_type=art_type,
                )
            else:
                uploaded_count += self._upload_files_to_s3(
                    prefix=prefix,
                    from_path=params.from_path,
                    artifact_type=art_type,
                )

        logger.info(f'Uploaded {uploaded_count} artifacts in {time.time() - start_time:.2f} seconds')

    def generate_presigned_json(
        self,
        *,
        commit_sha: t.Optional[str] = None,
        branch: t.Optional[str] = None,
        artifact_type: t.Optional[str] = None,
        folder: t.Optional[str] = None,
        expire_in_days: int = 4,
    ) -> t.Dict[str, str]:
        """Generate presigned URLs for artifacts in S3 storage.

        Generates presigned URLs for artifacts that would be uploaded to S3 storage. The
        URLs can be used to download the artifacts directly from S3.

        :param commit_sha: Optional commit SHA. If no commit_sha provided, will use 1)
            PIPELINE_COMMIT_SHA env var, 2) latest commit from branch
        :param branch: Optional Git branch. If no branch provided, will use current
            branch
        :param artifact_type: Type of artifacts to generate URLs for (debug, flash,
            metrics)
        :param folder: Base folder to generate relative paths from
        :param expire_in_days: Expiration time in days for the presigned URLs (default:
            4 days)

        :returns: Dictionary mapping relative paths to presigned URLs

        :raises S3Error: If S3 is not configured
        """
        params = ArtifactParams(
            commit_sha=commit_sha,
            branch=branch,
            folder=folder,
        )

        if not self.s3_client:
            raise S3Error('Configure S3 storage to generate presigned URLs')

        prefix = f'{self.settings.gitlab.project}/{params.commit_sha}/'
        s3_path = self._get_s3_path(prefix, params.from_path)

        def _get_presigned_url_task(_bucket: str, _obj_name: str) -> t.Tuple[str, str]:
            res = self.s3_client.get_presigned_url(  # type: ignore
                'GET',
                bucket_name=_bucket,
                object_name=_obj_name,
                expires=timedelta(days=expire_in_days),
            )
            if not res:
                raise S3Error(f'Failed to generate presigned URL for {_obj_name}')

            return _obj_name, res

        tasks = []
        presigned_urls: t.Dict[str, str] = {}
        bucket_zip_artifacts: t.Dict[str, t.Set[str]] = defaultdict(set)
        bucket_patterns: t.Dict[str, t.List[str]] = defaultdict(list)
        for art_type in self._get_artifact_types(artifact_type):
            config = self.settings.gitlab.artifacts.s3.configs[art_type]
            if config.zip_first:
                bucket_zip_artifacts[config.bucket].add(art_type)
            else:
                bucket_patterns[config.bucket].extend(self._get_patterns_for_type(art_type))

        for bucket, artifact_types in bucket_zip_artifacts.items():
            zip_filenames = {f'{art_type}.zip' for art_type in artifact_types}
            for obj in self.s3_client.list_objects(bucket, prefix=s3_path, recursive=True):
                output_path = Path(self.envs.IDF_PATH) / obj.object_name.replace(prefix, '')
                if output_path.name not in zip_filenames:
                    continue

                tasks.append(
                    lambda _bucket=bucket, _obj_name=obj.object_name: _get_presigned_url_task(_bucket, _obj_name)
                )

        for bucket, patterns in bucket_patterns.items():
            patterns_regexes = [
                re.compile(translate(pattern, recursive=True, include_hidden=True)) for pattern in patterns
            ]

            for obj in self.s3_client.list_objects(bucket, prefix=s3_path, recursive=True):
                output_path = Path(self.envs.IDF_PATH) / obj.object_name.replace(prefix, '')
                if not any(pattern.match(str(output_path)) for pattern in patterns_regexes):
                    continue

                tasks.append(
                    lambda _bucket=bucket, _obj_name=obj.object_name: _get_presigned_url_task(_bucket, _obj_name)
                )

        results = execute_concurrent_tasks(tasks, task_name='generating presigned URL')
        for obj_name, presigned_url in results:
            presigned_urls[obj_name.replace(prefix, '')] = presigned_url

        return presigned_urls

    def _download_presigned_json_from_pipeline(
        self, pipeline_id: str, presigned_json_filename: str = 'presigned.json'
    ) -> str:
        """Download presigned.json file from a specific GitLab pipeline.

        Uses a local cache to avoid re-downloading the same presigned.json file for the
        same pipeline ID.

        :param pipeline_id: GitLab pipeline ID to download presigned.json from
        :param presigned_json_filename: Name of the presigned.json file to download

        :returns: Path to the presigned.json file (cached or downloaded)

        :raises ArtifactError: If presigned.json cannot be found or downloaded
        """
        if not self.settings.gitlab.build_pipeline.presigned_json_job_name:
            raise ArtifactError('Presigned JSON job name is not configured')

        # Check cache first
        cache_dir = Path(tempfile.gettempdir()) / '.cache' / 'idf-ci' / 'presigned_json' / pipeline_id
        cached_file = cache_dir / presigned_json_filename

        if cached_file.exists():
            logger.info(f'Using cached {presigned_json_filename} for pipeline {pipeline_id}')
            return str(cached_file)

        logger.info(f'Downloading {presigned_json_filename} from pipeline {pipeline_id}')

        # Find the child pipeline with the configured name
        child_pipeline_id = None
        try:
            for bridge in self.project.pipelines.get(pipeline_id, lazy=True).bridges.list(iterator=True):
                if bridge.name == self.settings.gitlab.build_pipeline.workflow_name:
                    child_pipeline_id = bridge.downstream_pipeline['id']
                    break
        except Exception as e:
            raise ArtifactError(f'Failed to get child pipeline from pipeline {pipeline_id}: {e}')

        if not child_pipeline_id:
            raise ArtifactError(
                f'No child pipeline found for pipeline {pipeline_id} with name '
                f'{self.settings.gitlab.build_pipeline.workflow_name}'
            )

        # Get the child pipeline and find the job that generates presigned.json
        download_from_job = None
        try:
            for job in self.project.pipelines.get(child_pipeline_id, lazy=True).jobs.list(iterator=True):
                if job.name == self.settings.gitlab.build_pipeline.presigned_json_job_name:
                    download_from_job = job
                    break
        except Exception as e:
            raise ArtifactError(
                f'Failed to get job {self.settings.gitlab.build_pipeline.presigned_json_job_name} '
                f'from child pipeline {child_pipeline_id}: {e}'
            )

        if not download_from_job:
            raise ArtifactError(
                f'No job found in child pipeline {child_pipeline_id} with name '
                f'{self.settings.gitlab.build_pipeline.presigned_json_job_name}'
            )

        # Download the presigned.json file from the job artifacts
        try:
            artifact_data = self.project.jobs.get(download_from_job.id, lazy=True).artifact(presigned_json_filename)
        except Exception as e:
            raise ArtifactError(
                f'Failed to get artifact {presigned_json_filename} from job {download_from_job.id}: {e}'
            )

        # Create cache directory and save to cache
        cache_dir.mkdir(parents=True, exist_ok=True)
        with open(cached_file, 'wb') as fw:
            fw.write(artifact_data)

        logger.debug(f'Successfully downloaded and cached {presigned_json_filename} for pipeline {pipeline_id}')
        return str(cached_file)



# SPDX-FileCopyrightText: 2025-2026 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import json
import os
import textwrap
import typing as t
from pathlib import Path

import pytest
from bs4 import BeautifulSoup
from conftest import create_project
from esp_bool_parser.constants import SUPPORTED_TARGETS

from idf_ci.cli import click_cli


def match_app(data: dict, path: Path, target: str, sdkconfig: str, cb: t.Callable[[t.Dict[str, t.Any]], bool]) -> bool:
    projects = data.get('projects', {})
    normalized_path = Path(path).as_posix()
    project = None

    # Find project by path
    for project_path in projects:
        if Path(project_path).as_posix() == normalized_path:
            project = projects[project_path]
            break

    if project is None:
        return False

    for app in project.get('apps', []):
        if app.get('target') != target or app.get('sdkconfig') != sdkconfig:
            continue

        if cb(app):
            return True

    return False


def has_test_case(data: dict, path: Path, target: str, sdkconfig: str, test_case: str) -> bool:
    return match_app(
        data,
        path,
        target,
        sdkconfig,
        lambda app: any(tc.get('name') == test_case for tc in app.get('test_cases', [])),
    )


def has_test_case_by_caseid(data: dict, path: Path, target: str, sdkconfig: str, caseid: str) -> bool:
    return match_app(
        data,
        path,
        target,
        sdkconfig,
        lambda app: any(tc.get('caseid') == caseid for tc in app.get('test_cases', [])),
    )


def check_property(
    data: dict, path: Path, target: str, sdkconfig: str, property_name: str, property_value: t.Any
) -> bool:
    return match_app(
        data,
        path,
        target,
        sdkconfig,
        lambda app: app.get(property_name) == property_value,
    )


def check_test_case_property(
    data: dict, path: Path, target: str, sdkconfig: str, test_case: str, property_name: str, property_value: t.Any
) -> bool:
    return match_app(
        data,
        path,
        target,
        sdkconfig,
        lambda app: any(
            tc.get('name') == test_case and tc.get(property_name) == property_value for tc in app.get('test_cases', [])
        ),
    )


@pytest.mark.skipif(not os.getenv('IDF_PATH'), reason='IDF_PATH is not set')
class TestBuildCollect:
    @pytest.fixture(autouse=True)
    def setup_test_projects(self, tmp_path: Path):
        create_project('foo', tmp_path)
        create_project('bar', tmp_path)

        with open(tmp_path / '.idf_build_apps.toml', 'w') as fw:
            fw.write(
                textwrap.dedent(f"""
            config = [
                "sdkconfig.ci=default",
                "sdkconfig.ci.*=",
                "=default"
            ]

            check_manifest_rules = true
            manifest_rootpath = "{tmp_path.as_posix()}"
            manifest_filepatterns = [
                '**/.build-test-rules.yml',
            ]
            """)
            )

    def run_build_collect(
        self, root_path, runner, paths: t.Optional[t.List[Path]] = None, additional_args: t.Optional[t.List[str]] = None
    ) -> t.Dict[str, t.Any]:
        output_file = root_path / 'output.json'
        args = [
            'build',
            'collect',
            '--output',
            str(output_file),
        ]

        if paths is not None:
            for p in paths:
                args.extend(['--paths', str(p)])

        if additional_args is not None:
            args.extend(additional_args)

        result = runner.invoke(
            click_cli,
            args,
        )

        assert result.exit_code == 0
        assert output_file.exists()

        with open(output_file) as fr:
            data = json.load(fr)

        return data

    def test_build_collect_no_test_cases(self, tmp_path, runner) -> None:
        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases'] == 0

    def test_build_collect_with_test_cases(self, tmp_path, runner) -> None:
        with open(tmp_path / 'foo' / 'test_foo1.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('target', [
                'esp32',
            ], indirect=True)
            def test_foo(dut):
                pass
            """)
            )

        with open(tmp_path / 'bar' / 'test_bar1.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('target', [
                'esp32', 'esp32c3',
            ], indirect=True)
            def test_bar(dut):
                pass
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases_used'] == 3
        assert has_test_case(data, Path('foo'), 'esp32', 'default', 'test_foo')
        assert has_test_case(data, Path('bar'), 'esp32', 'default', 'test_bar')
        assert has_test_case(data, Path('bar'), 'esp32c3', 'default', 'test_bar')

    def test_build_collect_with_sdkconfig(self, tmp_path, runner) -> None:
        with open(tmp_path / 'foo' / 'test_foo2.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('target', [
                'esp32',
            ], indirect=True)
            @pytest.mark.parametrize('config', [
                'cfg1', 'cfg2',
            ])
            def test_foo(dut):
                pass
            """)
            )

        with open(tmp_path / 'foo' / 'sdkconfig.ci.cfg1', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            CONFIG_EXAMPLE_OPTION=y
            """)
            )

        with open(tmp_path / 'foo' / 'sdkconfig.ci.cfg2', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            CONFIG_EXAMPLE_OPTION=y
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases_used'] == 2
        assert has_test_case(data, Path('foo'), 'esp32', 'cfg1', 'test_foo')
        assert has_test_case(data, Path('foo'), 'esp32', 'cfg2', 'test_foo')

    def test_build_collect_include_disabled_apps(self, tmp_path, runner) -> None:
        with open(tmp_path / '.build-test-rules.yml', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            foo:
                disable:
                    - if: IDF_TARGET == "esp32"
                      reason: Disabled for esp32
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert check_property(
            data,
            Path('foo'),
            'esp32',
            'default',
            'build_status',
            'disabled',
        )

    def test_build_collect_supported_targets(self, tmp_path, runner) -> None:
        with open(tmp_path / 'foo' / 'test_foo3.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest
            from pytest_embedded_idf.utils import idf_parametrize

            @idf_parametrize('target', [
                ('supported_targets'),
            ], indirect=['target'])
            def test_foo(dut):
                pass
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases_used'] == len(SUPPORTED_TARGETS)

    def test_build_collect_test_comment(self, tmp_path, runner) -> None:
        with open(tmp_path / '.build-test-rules.yml', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            foo:
                disable:
                    - if: IDF_TARGET == "esp32"
                      reason: Disabled for esp32
                disable_test:
                    - if: IDF_TARGET == "esp32c3"
                      reason: Disabled test for esp32c3
            """)
            )

        data = self.run_build_collect(tmp_path, runner)

        # esp32
        assert check_property(
            data,
            Path('foo'),
            'esp32',
            'default',
            'build_comment',
            'Disabled by manifest rule: IDF_TARGET == "esp32" (reason: Disabled for esp32)',
        )

        assert check_property(
            data,
            Path('foo'),
            'esp32',
            'default',
            'test_comment',
            'Disabled by manifest rule: IDF_TARGET == "esp32" (reason: Disabled for esp32)',
        )

        # esp32c3
        assert check_property(
            data,
            Path('foo'),
            'esp32c3',
            'default',
            'build_comment',
            '',
        )

        assert check_property(
            data,
            Path('foo'),
            'esp32c3',
            'default',
            'test_comment',
            'Disabled by manifest rule: IDF_TARGET == "esp32c3" (reason: Disabled test for esp32c3)',
        )

    def test_build_collect_test_cases_requiring_nonexistent_app(self, tmp_path, runner) -> None:
        with open(tmp_path / 'foo' / 'test_foo4.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('config, target', [
                ('default', 'esp32'),
            ], indirect=True)
            def test_foo(dut):
                pass

            @pytest.mark.parametrize('config, target', [
                ('release', 'esp32')
            ], indirect=True)
            def test_bar(dut):
                pass
            """)
            )

        with open(tmp_path / 'foo' / 'sdkconfig.ci.release', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            CONFIG_EXAMPLE_OPTION=y
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases_used'] == 1
        assert data['summary']['total_test_cases_requiring_nonexistent_app'] == 1

    def test_build_collect_multiple_targets(self, tmp_path, runner) -> None:
        with open(tmp_path / 'foo' / 'test_foo5.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest
            from pytest_embedded_idf.utils import idf_parametrize

            @pytest.mark.generic_multi_device
            @pytest.mark.parametrize(
                'count, config',
                [
                    (2, 'release'),
                ],
                indirect=True,
            )
            @idf_parametrize('target', ['esp32'], indirect=['target'])
            def test_ledc_multi_device(case_tester) -> None:  # type: ignore
                case_tester.run_all_multi_dev_cases(reset=True)
            """)
            )

        with open(tmp_path / 'foo' / 'sdkconfig.ci.release', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            CONFIG_EXAMPLE_OPTION=y
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases_used'] == 1
        assert data['summary']['total_test_cases_requiring_nonexistent_app'] == 0

    def test_build_collect_include_host_test(self, tmp_path, runner) -> None:
        with open(tmp_path / 'foo' / 'test_foo6.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest
            from pytest_embedded_idf.utils import idf_parametrize

            @pytest.mark.host_test
            @idf_parametrize('target', ['esp32'], indirect=['target'])
            def test_host(dut):
                pass
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases_used'] == 1
        assert has_test_case(data, Path('foo'), 'esp32', 'default', 'test_host')

    def test_build_collect_include_only_enabled_apps(self, tmp_path, runner) -> None:
        with open(tmp_path / '.build-test-rules.yml', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            foo:
                disable:
                    - if: IDF_TARGET == "esp32"
                      reason: Disabled for esp32
            """)
            )

        with open(tmp_path / 'foo' / 'test_foo7.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest
            from pytest_embedded_idf.utils import idf_parametrize

            @idf_parametrize('target', ['esp32', 'esp32c3'], indirect=['target'])
            def test_foo(dut):
                pass
            """)
            )

        data = self.run_build_collect(tmp_path, runner, additional_args=['--include-only-enabled-apps'])
        assert data['summary']['total_test_cases_used'] == 1
        assert data['summary']['total_test_cases_requiring_nonexistent_app'] == 1
        assert has_test_case(data, Path('foo'), 'esp32c3', 'default', 'test_foo')
        assert not has_test_case(data, Path('foo'), 'esp32', 'default', 'test_foo')

    def test_build_collect_from_other_path(self, tmp_path, runner, monkeypatch) -> None:
        with open(tmp_path / 'foo' / 'test_foo8.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('target', [
                'esp32',
            ], indirect=True)
            def test_foo(dut):
                pass
            """)
            )

        other_path = tmp_path / 'baz'
        os.makedirs(other_path)
        monkeypatch.chdir(other_path)

        data = self.run_build_collect(other_path, runner, paths=[tmp_path / 'foo'])
        assert data['summary']['total_test_cases_used'] == 1
        assert has_test_case(data, (tmp_path / 'foo'), 'esp32', 'default', 'test_foo')

    def test_build_collect_temp_skip_ci_marker(self, tmp_path, runner) -> None:
        with open(tmp_path / 'foo' / 'test_foo9.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest

            @pytest.mark.parametrize('target', ['esp32', 'esp32c3'], indirect=True)
            @pytest.mark.temp_skip_ci(targets=['esp32c3'], reason='Temporary skip for CI')
            def test_foo1(dut):
                pass

            @pytest.mark.parametrize('target', ['esp32', 'esp32c3'], indirect=True)
            @pytest.mark.temp_skip(targets=['esp32c3'], reason='Temporary skip for CI')
            def test_foo2(dut):
                pass
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases'] == 4
        assert data['summary']['total_test_cases_used'] == 2
        assert data['summary']['total_test_cases_disabled'] == 2
        assert has_test_case(data, Path('foo'), 'esp32', 'default', 'test_foo1')
        assert has_test_case(data, Path('foo'), 'esp32', 'default', 'test_foo2')
        assert check_test_case_property(
            data,
            Path('foo'),
            'esp32c3',
            'default',
            'test_foo1',
            'disabled_by_marker',
            True,
        )
        assert check_test_case_property(
            data,
            Path('foo'),
            'esp32c3',
            'default',
            'test_foo1',
            'skip_reason',
            'Temporary skip for CI',
        )

    def test_build_collect_test_on_both_target_and_qemu(self, tmp_path, runner) -> None:
        with open(tmp_path / 'foo' / 'test_foo10.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest
            from pytest_embedded import Dut
            from pytest_embedded_idf.utils import idf_parametrize

            @idf_parametrize(
                'target,test_on,markers',
                [
                    ('esp32', 'target', (pytest.mark.generic)),
                    ('esp32', 'qemu', (pytest.mark.host_test, pytest.mark.qemu))
                ],
                indirect=['target']
            )
            def test_foo(dut: Dut, test_on: str) -> None:
                pass
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases'] == 2
        assert data['summary']['total_test_cases_used'] == 2
        assert has_test_case_by_caseid(data, Path('foo'), 'esp32', 'default', 'esp32.default.test_foo')
        assert has_test_case_by_caseid(data, Path('foo'), 'esp32', 'default', 'esp32_qemu.default.test_foo')

    def test_build_collect_test_disabled_by_manifest(self, tmp_path, runner) -> None:
        with open(tmp_path / '.build-test-rules.yml', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            foo:
                disable_test:
                    - if: IDF_TARGET == "esp32"
                      reason: Disabled test for esp32
            """)
            )

        with open(tmp_path / 'foo' / 'test_foo11.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
            import pytest
            from pytest_embedded_idf.utils import idf_parametrize

            @idf_parametrize('target', ['esp32', 'esp32c3'], indirect=['target'])
            def test_foo(dut):
                pass
            """)
            )

        data = self.run_build_collect(tmp_path, runner)
        assert data['summary']['total_test_cases'] == 2
        assert data['summary']['total_test_cases_used'] == 1
        assert data['summary']['total_test_cases_disabled'] == 1
        assert has_test_case(data, Path('foo'), 'esp32c3', 'default', 'test_foo')
        assert check_test_case_property(
            data,
            Path('foo'),
            'esp32',
            'default',
            'test_foo',
            'disabled_by_manifest',
            True,
        )
        assert check_test_case_property(
            data,
            Path('foo'),
            'esp32',
            'default',
            'test_foo',
            'test_comment',
            'Disabled by manifest rule: IDF_TARGET == "esp32" (reason: Disabled test for esp32)',
        )

    def test_build_collect_html_output(self, tmp_path, runner) -> None:
        output_file = tmp_path / 'output.html'

        args = [
            'build',
            'collect',
            '--format',
            'html',
            '--output',
            str(output_file),
        ]

        result = runner.invoke(
            click_cli,
            args,
        )

        assert result.exit_code == 0
        assert output_file.exists()
        with open(output_file) as fr:
            soup = BeautifulSoup(fr, 'html.parser')
            assert soup.find('html') is not None



# SPDX-FileCopyrightText: 2025-2026 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import os
import shutil
import sys
import textwrap

import minio
import pytest
import requests

from idf_ci.cli import click_cli
from idf_ci.idf_gitlab import ArtifactManager
from idf_ci.idf_gitlab.api import S3Error
from idf_ci.settings import _refresh_ci_settings


# to run this test, don't forget to run "docker compose up -d" in the root directory of the project
@pytest.mark.skipif(sys.platform == 'win32', reason='minio service not available on Windows')
class TestUploadDownloadArtifacts:
    @pytest.fixture
    def sample_artifacts_dir(self, tmp_path):
        build_dir = tmp_path / 'app' / 'build_esp32_build'
        build_dir.mkdir(parents=True)

        # Create some test files
        (build_dir / 'build.log').write_text('Test build log', encoding='utf-8')
        (build_dir / 'test.bin').write_text('Binary content', encoding='utf-8')
        (build_dir / 'size.json').write_text('{"size": 1024}', encoding='utf-8')
        (tmp_path / 'optional.txt').write_text('Optional content', encoding='utf-8')

        return build_dir

    @pytest.fixture(autouse=True)
    def setup_test_dir(self, tmp_path, tmp_dir, monkeypatch):
        (tmp_path / '.idf_ci.toml').write_text(
            textwrap.dedent("""
                [gitlab]
                project = "espressif/esp-idf"

                [gitlab.artifacts.s3]
                enable = true

                [gitlab.artifacts.s3.configs.debug]
                bucket = "private"
                zip_first = true
                base_dir_pattern = "**/build*/"
                patterns = ["build.log"]

                [gitlab.artifacts.s3.configs.flash]
                bucket = "private"
                zip_first = true
                base_dir_pattern = "**/build*/"
                patterns = ["*.bin"]

                [gitlab.artifacts.s3.configs.metrics]
                bucket = "private"
                base_dir_pattern = "**/build*/"
                patterns = ["size.json"]

                [gitlab.artifacts.s3.configs.optional]
                bucket = "public"
                is_public = true
                patterns = ["**/optional.txt"]
                if_clause = 'ENV_VAR_FOO == "foo"'
            """)
        )

        curdir = os.getcwd()
        os.chdir(tmp_path)
        _refresh_ci_settings()

        monkeypatch.setenv('IDF_S3_SERVER', 'http://localhost:9100')
        monkeypatch.setenv('IDF_S3_ACCESS_KEY', 'minioadmin')
        monkeypatch.setenv('IDF_S3_SECRET_KEY', 'minioadmin')

        monkeypatch.setenv('IDF_PATH', tmp_dir)

        yield

        os.chdir(curdir)

    @pytest.fixture
    def s3_client(self) -> minio.Minio:
        client = ArtifactManager().s3_client
        assert client is not None

        # remove all objects in both buckets before test
        for bucket in ['private', 'public']:
            for obj in client.list_objects(bucket, recursive=True):
                try:
                    client.remove_object(bucket, obj.object_name)
                except minio.error.S3Error as e:
                    logging.error(f'Error removing object: {e}')

        return client

    def test_all_without_optional(self, runner, s3_client, sample_artifacts_dir):
        commit_sha = 'cli_test_sha_123'

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                commit_sha,
            ],
        )
        assert result.exit_code == 0

        objs = list(s3_client.list_objects('private', recursive=True))
        assert len(objs) == 3
        assert sorted(obj.object_name for obj in objs) == [
            f'espressif/esp-idf/{commit_sha}/app/build_esp32_build/debug.zip',
            f'espressif/esp-idf/{commit_sha}/app/build_esp32_build/flash.zip',
            f'espressif/esp-idf/{commit_sha}/app/build_esp32_build/size.json',
        ]

        shutil.rmtree(sample_artifacts_dir)

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'download-artifacts',
                '--commit-sha',
                commit_sha,
            ],
        )
        assert result.exit_code == 0

        assert sorted(os.listdir(sample_artifacts_dir)) == [
            'build.log',
            'size.json',
            'test.bin',
        ]

        # test single artifact type download
        shutil.rmtree(sample_artifacts_dir)

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'download-artifacts',
                '--commit-sha',
                commit_sha,
                '--type',
                'flash',
            ],
        )
        assert result.exit_code == 0
        assert sorted(os.listdir(sample_artifacts_dir)) == [
            'test.bin',
        ]

    def test_all_with_optional(self, runner, s3_client, sample_artifacts_dir, tmp_path, monkeypatch):
        commit_sha = 'cli_test_sha_123'

        monkeypatch.setenv('ENV_VAR_FOO', 'foo')

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                commit_sha,
            ],
        )
        assert result.exit_code == 0
        private_objs = list(s3_client.list_objects('private', recursive=True))
        public_objs = list(s3_client.list_objects('public', recursive=True))

        assert len(private_objs) == 3
        assert sorted(obj.object_name for obj in private_objs) == [
            f'espressif/esp-idf/{commit_sha}/app/build_esp32_build/debug.zip',
            f'espressif/esp-idf/{commit_sha}/app/build_esp32_build/flash.zip',
            f'espressif/esp-idf/{commit_sha}/app/build_esp32_build/size.json',
        ]

        assert len(public_objs) == 1
        assert public_objs[0].object_name == f'espressif/esp-idf/{commit_sha}/optional.txt'

        shutil.rmtree(sample_artifacts_dir)
        (tmp_path / 'optional.txt').unlink(missing_ok=True)

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'download-artifacts',
                '--commit-sha',
                commit_sha,
            ],
        )
        assert result.exit_code == 0
        assert sorted(os.listdir(sample_artifacts_dir)) == [
            'build.log',
            'size.json',
            'test.bin',
        ]
        assert (tmp_path / 'optional.txt').exists()

    def test_cli_generate_presigned_json(self, runner):
        commit_sha = 'cli_test_sha_123'

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                commit_sha,
                '--type',
                'metrics',
            ],
        )
        assert result.exit_code == 0

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'generate-presigned-json',
                '--commit-sha',
                commit_sha,
                '--type',
                'metrics',
            ],
        )
        assert result.exit_code == 0

        presigned_urls = json.loads(result.stdout)
        assert sorted(presigned_urls.keys()) == ['app/build_esp32_build/size.json']

        response = requests.get(presigned_urls['app/build_esp32_build/size.json'])
        assert response.status_code == 200

    def test_cli_download_with_presigned_json(self, runner, tmp_path, sample_artifacts_dir):
        commit_sha = 'presigned_test_sha_123'

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                commit_sha,
                '--type',
                'flash',
            ],
        )
        assert result.exit_code == 0

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                commit_sha,
                '--type',
                'metrics',
            ],
        )
        assert result.exit_code == 0

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'generate-presigned-json',
                '--commit-sha',
                commit_sha,
                '-o',
                str(tmp_path / 'presigned.json'),
            ],
        )
        assert result.exit_code == 0

        assert (tmp_path / 'presigned.json').exists()
        with open('presigned.json') as fr:
            assert sorted(json.load(fr).keys()) == [
                'app/build_esp32_build/flash.zip',
                'app/build_esp32_build/size.json',
            ]

        shutil.rmtree(sample_artifacts_dir, ignore_errors=True)

        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'download-artifacts',
                '--commit-sha',
                commit_sha,
                '--presigned-json',
                str(tmp_path / 'presigned.json'),
            ],
        )
        assert result.exit_code == 0

        # Verify the file was downloaded and extracted
        assert (sample_artifacts_dir / 'test.bin').exists()
        assert (sample_artifacts_dir / 'test.bin').read_text() == 'Binary content'
        assert (sample_artifacts_dir / 'size.json').exists()
        assert (sample_artifacts_dir / 'size.json').read_text() == '{"size": 1024}'

    # Error Handling Tests
    def test_download_without_s3_credentials(self, runner, tmp_path, monkeypatch):
        # Remove S3 credentials
        monkeypatch.delenv('IDF_S3_ACCESS_KEY')

        # Try to download artifacts
        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'download-artifacts',
                '--commit-sha',
                'test_sha',
                '--type',
                'flash',
                str(tmp_path),
            ],
        )
        assert result.exit_code != 0
        assert isinstance(result.exception, S3Error)
        assert 'S3 client is not configured properly' in result.exception.args[0]

    def test_upload_without_s3_credentials(
        self,
        runner,
        tmp_path,
        sample_artifacts_dir,  # noqa
        monkeypatch,
    ):
        # Remove S3 credentials
        monkeypatch.delenv('IDF_S3_ACCESS_KEY')

        # Try to upload artifacts
        result = runner.invoke(
            click_cli,
            [
                'gitlab',
                'upload-artifacts',
                '--commit-sha',
                'test_sha',
                '--type',
                'flash',
                str(tmp_path),
            ],
        )

        assert result.exit_code != 0
        assert isinstance(result.exception, S3Error)
        assert 'Configure S3 storage to upload artifacts' in result.exception.args[0]
